package boundingboxeditor.model.io;

import boundingboxeditor.model.data.BoundingBoxData;
import boundingboxeditor.model.data.BoundingShapeData;
import boundingboxeditor.model.data.ImageAnnotation;
import boundingboxeditor.model.data.ImageAnnotationData;
import boundingboxeditor.model.io.results.IOErrorInfoEntry;
import boundingboxeditor.model.io.results.ImageAnnotationExportResult;
import javafx.beans.property.DoubleProperty;
import javafx.geometry.Bounds;
import org.apache.commons.lang3.StringUtils;

import java.io.BufferedWriter;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;

/**
 * Saves rectangular bounding-box annotations in the YOLO-format described at
 * https://github.com/AlexeyAB/Yolo_mark/issues/60#issuecomment-401854885
 */
public class YOLOSaveStrategy implements ImageAnnotationSaveStrategy {
    private static final DecimalFormat DECIMAL_FORMAT =
            new DecimalFormat("#.######", DecimalFormatSymbols.getInstance(Locale.ENGLISH));
    private static final String YOLO_ANNOTATION_FILE_EXTENSION = ".txt";
    private static final String OBJECT_DATA_FILE_NAME = "object.data";
    private Path saveFolderPath;
    private List<String> categories;

    @Override
    public ImageAnnotationExportResult save(ImageAnnotationData annotations, Path destination,
                                            DoubleProperty progress) {
        this.saveFolderPath = destination;
        this.categories = annotations.getCategoryNameToBoundingShapeCountMap().entrySet().stream()
                                     .filter(stringIntegerEntry -> stringIntegerEntry.getValue() > 0)
                                     .map(Map.Entry::getKey)
                                     .sorted()
                                     .collect(Collectors.toList());

        List<IOErrorInfoEntry> unParsedFileErrorMessages = Collections.synchronizedList(new ArrayList<>());

        try {
            createObjectDataFile();
        } catch(IOException e) {
            unParsedFileErrorMessages.add(new IOErrorInfoEntry(OBJECT_DATA_FILE_NAME, e.getMessage()));
        }

        int totalNrOfAnnotations = annotations.getImageAnnotations().size();
        AtomicInteger nrProcessedAnnotations = new AtomicInteger(0);

        annotations.getImageAnnotations().parallelStream().forEach(annotation -> {
            try {
                createAnnotationFile(annotation);
            } catch(IOException e) {
                unParsedFileErrorMessages
                        .add(new IOErrorInfoEntry(annotation.getImageFileName(), e.getMessage()));
            }

            progress.set(1.0 * nrProcessedAnnotations.incrementAndGet() / totalNrOfAnnotations);
        });

        return new ImageAnnotationExportResult(
                totalNrOfAnnotations - unParsedFileErrorMessages.size(),
                unParsedFileErrorMessages
        );
    }

    private void createObjectDataFile() throws IOException {
        try(BufferedWriter fileWriter = Files.newBufferedWriter(saveFolderPath
                                                                        .resolve(OBJECT_DATA_FILE_NAME))) {
            for(int i = 0; i < categories.size() - 1; ++i) {
                fileWriter.write(categories.get(i));
                fileWriter.newLine();
            }

            if(!categories.isEmpty()) {
                fileWriter.write(categories.get(categories.size() - 1));
            }
        }
    }

    private void createAnnotationFile(ImageAnnotation annotation) throws IOException {
        String imageFileName = annotation.getImageFileName();
        String imageFileNameWithoutExtension = imageFileName.substring(0, imageFileName.lastIndexOf('.'));

        try(BufferedWriter fileWriter = Files.newBufferedWriter(saveFolderPath
                                                                        .resolve(imageFileNameWithoutExtension +
                                                                                         YOLO_ANNOTATION_FILE_EXTENSION))) {
            List<BoundingShapeData> boundingShapeDataList = annotation.getBoundingShapeData();

            for(int i = 0; i < boundingShapeDataList.size() - 1; ++i) {
                BoundingShapeData boundingShapeData = boundingShapeDataList.get(i);

                if(boundingShapeData instanceof BoundingBoxData) {
                    fileWriter.write(createBoundingBoxDataEntry((BoundingBoxData) boundingShapeData));
                    fileWriter.newLine();
                }
            }

            if(!boundingShapeDataList.isEmpty()) {
                BoundingShapeData lastShapeData = boundingShapeDataList.get(boundingShapeDataList.size() - 1);

                if(lastShapeData instanceof BoundingBoxData) {
                    fileWriter.write(createBoundingBoxDataEntry((BoundingBoxData) lastShapeData));
                }
            }
        }
    }

    private String createBoundingBoxDataEntry(BoundingBoxData boundingBoxData) {
        int categoryIndex = categories.indexOf(boundingBoxData.getCategoryName());

        Bounds relativeBounds = boundingBoxData.getRelativeBoundsInImage();

        String xMidRelative = DECIMAL_FORMAT.format(relativeBounds.getCenterX());
        String yMidRelative = DECIMAL_FORMAT.format(relativeBounds.getCenterY());
        String widthRelative = DECIMAL_FORMAT.format(relativeBounds.getWidth());
        String heightRelative = DECIMAL_FORMAT.format(relativeBounds.getHeight());

        return StringUtils.join(List.of(categoryIndex, xMidRelative, yMidRelative, widthRelative, heightRelative), " ");
    }
}
